/* input.c

   file and character expansion functions...
*/

#include "vpt.h"
#include <sys/stat.h>

#define MAXPATHS 20
#define MAXLINE 300
#define MAXPARAMS 10

static u_char	pathcount;

static FILE	*paths[MAXPATHS],
		*curpath,
		*auxinfile;

static u_char	linebuff[MAXLINE],
		*lineptr;
				
static char	*filenames[MAXPATHS];
static u_int	linecounters[MAXPATHS];

static char paramcount;
static struct paramstruct
{
	short p_maclevel, p_ncount;
	KCHAR p_odelim, p_cdelim;
}paramstack[MAXPARAMS];

#define MAXREPEATS 10

static int repeatlv;	 /* current repeat level */
static struct
{
	FILE *rp_path;		// path  when repeat started
	int rp_ptr;		// file pos when repeat started
	int rp_count;		// number of times left to repeat
	u_int linecounters;	// value of linecounter at start of repeat
}
repeats[MAXREPEATS];

static char labelsearch;  /* signals label search in progress */

static void startpm(int mlevel);
static KCHAR get_c(void);
static KCHAR getcinput(void);
static void icommnt(void);
static void gethcode(void);
static void getaux(void);
static void getshvar(void);
static KCHAR gotogetkey(void);


/*	============================================================
	Include another input file (SI)
*/

void
switchin(void)
{
	char filename[300];
	
	skipequal();
	
	getstring(filename, sizeof(filename)-2, "filename too");
	
	if(!newfile(filename)) error("Can't open '%s' for input", filename);
}


/*	============================================================
	Open another input file. The filepointers are kept on a stack.
   	
	This function will return 0 if the file cannot be opened.
*/

FILE *
newfile(char *s)
{
	FILE *f;
	char *p;
	
	if(pathcount>=MAXPATHS-1) terminate("Too many nested files");
	
	if(strcmp(s, "@")==0)
	{
		f=stdin;
		s="STDIN";
	}
	else f=fopen(s, "r");
	
	if(f)
	{
		paths[pathcount]=curpath=f;
		if((p=rindex(s, '/'))) p++;  /* point to actual file NAME */
		else p=s;
		if(filenames[pathcount]) free(filenames[pathcount]);
		filenames[pathcount]=strdup(p);
		if(!pathcount) add_macro_def("FILE", p);
		setfmacs(paths[pathcount], filenames[pathcount]);
		pathcount++;
		lineptr=NULL;  /* signal empty input buffer */
	}
	return f;
}


/*	============================================================
	See if the next character is a comma. If it is, return 1.
	If not a comma, return the char to input and return 0.
	
	This is used by the command parser for command options.
*/

int
comma_ck(void)
{
	KCHAR c=getnonb();
	
	if(c==',') return 1;
	
	returnkey(c);
	return 0;
}

/*	============================================================
	Skip over an equal sign and any spaces
*/

void
skipequal(void)
{
	KCHAR c;
	
	if((c=getnonb())!='=') returnkey(c);
}

/*	============================================================
	Get a non-space character
*/

KCHAR
getnonb(void)
{
	KCHAR c;
	
	while((c=getkey())==SPACE);
	
	return c;
}

/*	============================================================
	Switch input for parm. We maintain a stack which contains the
	macro nesting level when the param was started and the terminating
	delimiter for the param.
*/

static void
startpm(int mlevel)
{
	int t=paramcount;
	KCHAR c;
	struct paramstruct *ps;
	
	if(mlevel==0) error("\\! only permitted in macros");
	
	if(t>=MAXPARAMS) terminate("Too many nested macro parameters");
	
	ps=&paramstack[t];
	paramcount++; 	/* do this BEFORE getting the param delimiter */
			/* but leave 't' pointing to the correct slot */
	
	/*	we get the next key from the input stream, not the macro.
		Since the paramcount has been incremented, this will be the
		char. following the macro. Note that we zap the delims for this
		so we don't accidently dump out of the param....
	*/
	
	ps->p_maclevel=maclevel;
	ps->p_odelim=ps->p_cdelim=ps->p_ncount=0;
	c=getnonb();
	ps->p_odelim=c;
	c=isdelm(c);
	if(c==0) error("Illegal delimiter for parameter");
	ps->p_cdelim=c;
}


/*	============================================================
	Get a character from the current input stream or macro...
*/

KCHAR
getkey(void)
{
	KCHAR c1, c2, c3;
	int mlevel;
	
	for( ; ; )
	{
		c1=get_c();	   /* get char from buffer/input stream */
		
		if(!doingcmd && hardspace && c1==hardspace)
		{
			c1=INT_HSPACE;
			break;
		}
		
		if(!doingcmd && ghosthy && c1==ghosthy)
		{
			c1=INT_GH;
			break;
		}
		
		if(incomment || c1!=control) break;  /* not control, just return */
		
		mlevel=maclevel;
		c1=get_c();
		if(c1==control) break;	 /* char was a '\\', return '\' */
		
		if(ghosthy && c1==ghosthy) break;	// escaped ghost hypen, return char
		
		if(hardspace && c1==hardspace) break;	// escaped hardspace, return char
		
		
		/*	do macro expansion and tokenization.
			
			All the expansion/tokenize routines push their stuff onto
			the input stack so that when we do the continue we can go
			back and get the result at the top of this function.
			
			This first switch covers un-ambiguous single character cmds.
			None of these can be the start of a 2char sequence or an
			attribute.
		*/
		
		switch(f_tolower(c1))
		{
			case CR:
				returnkey(SPACE);
				continue;
				
			case '!':
				startpm(mlevel);	/* switch input for parm */
				continue;
				
			case '#':
				expnum();	 /* expand number		  */
				continue;
				
			case '@':
			case '$':
				expmac();	 /* expand macro		  */
				continue;
				
			case '%':
				tokdkey();	/* tokenize def. key	 */
				continue;
				
			case '*':
				icommnt();	  /* imbedded comment	   */
				continue;
				
			case '?':
				getaux();	 /* start input from aux  */
				continue;
				
			case 'h':
			case 'H':
				gethcode();	   /* 2 digit hex			*/
				continue;
				
			default:
				break;		/* drop to 2 char switch */
		}
		
		/*	Possibly a 2 character sequence so we get the next char. 
			1st check to see if an attribute...
		*/
		
		c2=get_c();
		if(f_isalnum(c1) && (c2=='+' || c2=='-' || c2=='*'))
		{
			tokattr(c1, c2);
			continue;
		}
		
		/* somewhat ambiguous 2 char commands */
		
		switch((f_tolower(c1)<<8)+f_tolower(c2))
		{
			dcase('c', 'b'):
				setcbarstatus();
				continue;
				
			dcase('c', 'f'):
				returnkey(ascregnum(getbodyfont()));
				continue;
				
			dcase('c', 'r'):
				returnkey((KCHAR)INT_CR);
				continue;
				
			dcase('e', 'v'):
				getshvar();
				continue;
				
			dcase('i', 'd'):
				tokdelayid();
				continue;
				
			dcase('m', 'n'):
				getmarnote();
				continue;
				
			dcase('t', 's'):
				toksettab();
				continue;
				
			dcase('t', 'w'):
				returnkey((KCHAR)INT_WIDETAB);
				continue;
				
			dcase('t', 'b'):
				returnkey((KCHAR)INT_COLTAB);
				continue;
			
			dcase('l', 'e'):	/* left$() & len()	*/
			dcase('l', 'o'):	/* low$()		*/
			dcase('m', 'i'):	/* mid$()		*/
			dcase('r', 'i'):	/* right$()		*/
			dcase('s', 't'):	/* string$() & strchr()	*/
			dcase('t', 'r'):	/* trim$()		*/
			dcase('u', 'p'):	/* up$()		*/
			dcase('v', 'a'):	/* val()		*/
				embeddedstr(c1, c2);
				continue;
			
			dcase('p', 'm'):	/* pad mode on/off +/- */
				c3=get_c();
				if(c3=='-')
				{
					returnkey((KCHAR)INT_PAD_OFF);
					continue;
				}
				if(c3=='+')
				{
					returnkey((KCHAR)INT_PAD_ON);
					continue;
				}
				returnkey(c3); break;
			
			default:
				break;
		}
		
		/* remaining 1 char commands, all take 2nd specifier or delim */
		
		returnkey(c2);
		switch(f_tolower(c1))
		{
			case 'a':
				expaux();	 	/* send to aux file	*/
				continue;
				
			case 'c':
				exptoc(); 	 	/* send to toc file	*/
				continue;
				
			case 'x':
				expinx();		/* send to index file	*/
				continue;
			
			case 'f':
				tokfont();		/* tokenize font mark	*/
				continue;
				
			case 'i':
				tokidtab();		/* tokenize delayid+tab */
				continue;
				
			case 'n':
				returnkey((KCHAR)INT_NOTE);
				continue;
				
			case 't':
				toktab();		/* tokenize tab		*/
				continue;
				
			case 'w':
				tokwait();		/* tokenize wait cmd	*/
				continue;
				
			default:
				break;		/* drop to unknown error() */
		}
		error("Unknown embedded sequence");
	}
	return c1;
}

static KCHAR
get_c(void)
{
	KCHAR c;
	struct paramstruct *ps;

loop:
	if((c=getretkey())) goto exit;   /* got returned key */
	
	/*	if paramcount and the current macro level are the same
		then we are filling in a param, so we get the key from
		input instead of the macro. If the param invokes a macro
		this still works since the levels will not match.
	*/
	
	if(paramcount)
	{
		ps=&paramstack[paramcount-1];
		if(ps->p_maclevel==maclevel)
		{
			c=getcinput();
			if(c==ps->p_odelim)
			{
				ps->p_ncount++;
				goto exit;
			}
			if(c==ps->p_cdelim)
			{
				if(ps->p_ncount==0)
				{
					paramcount--;
					goto loop;
				}
				ps->p_ncount--;
				goto exit;
			}
			if(!c) terminate("End of file encountered while processing macro parameter");
			goto exit;
		}
	}
	
	/*	try to get a key from a macro, if a NULL is returned
		then we get one from the current input. Note that getting a
		NULL from input is okay since this signals an EOF condition.
	*/
	
	if((c=getcmac())) goto exit;
	
	c=getcinput();
	
exit: 
	return (KCHAR) c;
}

/*	======================================================================
	read from input file. We read one line at a time. If the line
	read call in not successful we drop the pathcount to the parent
	file. If no parent we return a 0
*/

static KCHAR
getcinput(void)
{
	KCHAR c, t;
	
	for( ; ; )
	{
		c=lineptr ? *lineptr++ : 0;
		
		if(c==CR) linecounters[pathcount-1]++;
		
		if(!c)
		{
			if(curpath)
			{
				lineptr=(u_char*)fgets(linebuff, MAXLINE-1, curpath);
				if(lineptr) continue;
			}
			
			/* we get here if we can't read another line */
			
			if(labelsearch) break;  /* don't close file, gotogetkey */
						/* will do a rewind or terminate */
			
			/* close existing file and dec pathcount. If file is stdin
			   then leave the path open (incase it gets reopened later;
			   however we still dec the count. If the file being close
			   has any repeats pending we do an implicit re. This makes
			   it easy to include a repeat count on the command line.
			*/
			
			if(repeatlv && repeats[repeatlv-1].rp_path==curpath)
			{
				repeatend();
				getkey();  /* discard CR pushed by repeatend() */
				continue;
			}
			
			if(curpath && fileno(curpath)>2) fclose(curpath);
			
			for(t=0; t<repeatlv; )
			{
				if(repeats[t++].rp_path==curpath)
					terminate("File closed while repeat pending");
			}
			
			if(pathcount>1)			 /* backup to prior file */
			{
				if(paramcount)
					error("End of file reached with unterminated macro parameter");
				pathcount--;
				lineptr=NULL;											setfmacs(paths[pathcount-1], filenames[pathcount-1]);
				curpath=paths[pathcount-1];
				paramcount=0;
				continue;
			}
			else
			{
				curpath=0;	 /* this was last, return the null */
				lineptr=NULL;
				c=0;
				break;
			}
		}
		if(c=='\t')
		{
			c=INT_COLTAB;
			break;
		}
		
		if(c==CR || (c>=SPACE && c<=0xff)) break;  /* valid char, exit */
	}
	return c;
}


/*	============================================================
	Interpret the next two chars in the input stream as a hex
	value.... \h12, \HA2, etc. Note that this must be a 2digit string!!
*/

static void
gethcode(void)
{
	KCHAR v, c;
	char t;
	
	for(t=v=0; t<2; t++)
	{
		if(!f_isxdigit(c=getkey()))
			terminate("Two hex digits must follow a '\\h'");
		v=(v<<4) + ((c>'9') ? f_toupper(c)-'A'+10 : c-'0');
	}
	if((v<SPACE && v!=CR) || v>0xff)
		error("The value $%02x is not permitted as an embedded code", v);
	
	returnkey(v);
}

/*	============================================================
	Imbedded comment
*/

static void
icommnt(void)
{
	KCHAR delim;
	KCHAR c;
	
	incomment++;
	delim=getdelm();
	while((c=getkey()) && c!=delim);
	incomment--;
}


/*	============================================================
	Print the current input line if an error occurs. Called from
	terminate(). Also cancels the current input line--incase we
	are continuing after the error...
*/

void
errlineout(void)
{
	if(*linebuff) fprintf(stderr, "Input line: %s\n", linebuff);
	*linebuff='\0';
}


/*	============================================================
	Get a input line  from aux input file. If the file is no open we generate
	an error. Note that the trailing CR is striped from the input. Once
	the line is read it is put back on the input stack for interpretation.
	We pretend that this is a comment since the reading had no effect on
	controls....
	
	Remember, \? reads a single LINE from the file opened by S?. The way
	the code is right now a line must be no longer than 1024 bytes. If you
	want CRs in the data, you have to use \CR notation in the aux data file.
*/

#define AUXBUFF 1024

static void
getaux(void)
{
	static u_char *buff=NULL;
	u_char c, *p1, *p2;
	FLAG tty=isatty(fileno(auxinfile));	// set if input file is a tty
	
	if(!auxinfile) terminate("Auxillary input file not open");
	
	if(buff==NULL)
	{
		buff=(u_char *)malloc(AUXBUFF);
		if(buff==NULL)
			terminate("Unable to allocate buffer for auxillary input file");
	}
	
	if((fgets(buff, AUXBUFF-2, auxinfile))==NULL)
	{
		if(fileno(auxinfile)>2) fclose(auxinfile);
		auxinfile=0;
		terminate("EOF encountered on aux input file");
	}
	
	/* parse input line and strip CRs and other controls */
	
	for(p1=p2=buff; ; )
	{
		if((c=*p1++)==CR || !c)
		{
			*p2='\0';
			break;
		}
		
		if(tty && c==0x08)
		{
			if(p2>buff) p2--;
			continue;
		}
		
		if(c>=' ' && c<=0xff) *p2++=c;
	}
	
	/* return to input stream */
	
	addmacro(buff);
	
#undef AUXBUFF
}


/*	============================================================
	Open an auxillary input file (S?=filename)
*/

void
openauxinp(void)
{
	char filename[255];
	KCHAR c;
	
	if(auxinfile && fileno(auxinfile)>2)
	{
		fclose(auxinfile);
		auxinfile=0;
	}
	
	skipequal();
	if((c=getnonb())==CR || c==';' || c==0) *filename='\0';
	else
	{
		returnkey(c);
		getstring(filename, sizeof(filename)-2,
					"aux input filename");
	}
	
	if(strcmp(filename, "@")==0 || *filename=='\0') auxinfile=stdin;
	
	else
	{
		if((auxinfile=fopen(filename, "r"))==NULL)
			terminate("Can't open '%s' for aux input", filename);
	}
}

/*	============================================================
	Repeat. This maintains a stack of pointers for repeats. Repeats
	must start/end in the same file. When a 're' is found we check
	to see if any repeats are pending, and make sure that this repeat
	is in the same file. If it is, we backup to the saved position in
	the file.
*/

void
repeatstart(void)
{
	if(repeatlv>=MAXREPEATS) terminate("Too many repeats");
	
	if(!isrbf(curpath)) terminate("Repeats are only premitted in RBF files");
	
	skipequal();
	if((repeats[repeatlv].rp_count=getval(0))!=0)	  /* ignore rp=0	  */
	{
		repeats[repeatlv].rp_path=curpath;		/* current path		*/
		repeats[repeatlv].rp_ptr=ftell(curpath);	/* current file pos */
		repeats[repeatlv].linecounters=linecounters[pathcount-1];
		repeatlv++;
	}
}

void
repeatend(void)
{
	int t=repeatlv;
	
	if(!t--) error("Repeat end without start");
	
	if(repeats[t].rp_path != curpath)
		terminate("File changed from repeat start");
	
	if(--repeats[t].rp_count == 0) repeatlv--;  /* last repeat, exit */
	else
	{
		fseek(curpath, repeats[t].rp_ptr, 0);
		linecounters[pathcount-1]=repeats[t].linecounters;
		lineptr=NULL;				/* discard rest of this line */		returnkey(CR);				/* so cmd knows it's done */
	}
}


int
getrepeatcount(void)
{
	return (!repeatlv) ? 0 : repeats[repeatlv-1].rp_count;
}


/*	============================================================
	This is called by input after a \ev embedded command. We have
	to parse off the name following the command, find the shell
	varaible it refers to and return that to the input stream.
   	
	Eg. The value of the PATH var is \ev[PATH]
   	
	should print as
	
	   The value of the PATH var /dd/cmds/procfiles
*/

static void
getshvar(void)
{
	char varname[100], *p;
	extern char  *getenv(const char *);
	
	getstring(varname, sizeof(varname)-2, "shell variable name");
	if((p=getenv(varname))!=NULL)  returnstr(p);
}

/*	============================================================
	Handle a goto. This is complicated by the following:
   	
	1. If a macro is being executed the macro is checked for the label,
	2. The current file is checked from the current pos to EOF for the label,
	3. The current file is checked from the start to the current pos.
	
	If all three fail, an error is generated.
   	
	Note that labels must start a command line (blanks spaces are skipped).
*/

static char pass;

void
dogoto(void)
{
	char label[100];
	char *p;
	KCHAR c;
	
	skipequal();
	getstring(label, sizeof(label)-2, "GT label");
	if(!*label) terminate("Label expected after GT");
	
	if(gotomacro(label)) goto exit;	  /* check macro !!! */
	
	if(!isrbf(curpath))
		terminate("GOTO/LABEL must be in macro in non-RBF file");
	
	/*	check the file. gotogetkey() will take care of rewinding the file
		when we get to the end without a match.
	*/
	
	incomment++;
	labelsearch++;
	pass=0;
	
	for( ; ; )
	{
		while((c=gotogetkey())!=CR);	/* find a CR				*/
		if((c=gotogetkey())!=dot)  	/* followed by a '.'		*/
		{
			returnkey(c);		/* not '.', might be CR		*/
			continue;		/* so send it back and loop */
		}
		while((c=gotogetkey())==' '); 	/* skip leading blanks */
		if(c!=':')			/* label start?		   */
		{
			returnkey(c);		/* nope, loop		   */
			continue;
		}
		
		/* found a label, see if it matches target */
		
		p=label;
		while((f_tolower(c=gotogetkey()))==f_tolower(*p++));
		
		if(!*(--p))		  	/* did we reach end of target? */
		{
			incomment--;	/* yes! it's a match, so we	   */
			labelsearch--;  /* clean up and continue doing */
			returnkey(c); 	/* the file...				   */
			break;
		}
		returnkey(c);
	}
exit:;
}

static KCHAR
gotogetkey(void)
{
	KCHAR c;
	
	for( ; ; )
	{
		if((c=getkey())!=0) break;	/* got key, return */
		
		if(pass==0)				/* doing pass 1, rewind file */
		{
			fseek(curpath, 0, 0);
			linecounters[pathcount-1]=0;
			pass++;
			continue;
		}
		
		incomment--;			 /* checked entire file, error */
		labelsearch--;
		terminate("Label not found");
	}
	return c;
}



/*	============================================================
	See if the specified file is an regular file (ie. not a pipe, etc)
*/

int
isrbf(FILE *fpath)
{
	struct stat sbuff;
	
	if((fstat(fileno(fpath), &sbuff))!=0) return 0;	// error, assume not a reg. file
	
	return S_ISREG(sbuff.st_mode);		// return TRUE if regular file
}


/*	============================================================
	Return the current line number and filename for error and macro define
*/

u_int
get_current_linenumber(void)
{
	return linecounters[pathcount-1];
}

char *
get_current_filename(void)
{
	return filenames[pathcount-1];
}
