/* math.c  */

#include "vpt.h"
#include "att_table.h"
#include <ctype.h>

#define mult10(x) ((x<<3)+(x<<1))

static u_int parseval(char msflag);
static u_int strsize(void);


/*	============================================================
	Get a byte value.
	
	The value must be greater or equal to the min param. This is
	called mostly from cmd.c in setting values.
	
	If value is out of range an "illegal value" error is done
*/

u_int
getBval(int min)
{
	u_int t;
	
	t=getval(0);
	if(t<min || t>0xff) error("Illegal value '%d'", t);
	
	return t;
}

u_int
getAval(void)
{
	u_int t;

	t=getBval(0);
	if(t<' ' && t!=0) error("Illegal value '%d'", t);
	return t;
}

/*	============================================================
	Evaluate an expression.
   	
	This does a strict left to right parse of the expression (which
	must be gotten from the current input stream). All math is 32bit
	unsigned.
   	
   	msflag should be set if the result needs to be convert to microspaces
   	
   	operators:	 +, -, /, *, %
   	
   	constants: nnn - decimal number
			   $nn - hex number
			   'c  - character constant
			   "abc" - length of string in microspaces
			   MS  - microspaces per 10th inch
			   
	variables: #n - contents of number reg 'n'
			   LC - current line
			   LP - lines per page
			   MT - top margin
			   MB - bottom margin
			   RP - repeat count
			   CO - current number of cols
			   OE - last os9 error code
			   TF - tab fill value
			   CC - control char
			   CL - control line char
			   HS - hard space char
			   GH - ghost hypen char
			   JM - justify mode
			   LF - line format (main body)
			   MN - active margin for maginal notes
					(0-none,1-left,2-rite,3=inner,4=outer)
			   
			   all these are stored in microspaces
			   
			   IL - left indent
			   IR - right indent
			   ID - delayed indent
			   IT - temp indent
			   IC - centered indent
			   IB - both sides indent
			   ML[e/o] - left margin
			   MR[e/o] - right margin
			   t0..t9 - tab settings
*/


u_int
getval(char msflag)
{
	u_int accum;
	KCHAR op;
	
	accum=parseval(msflag);
	for( ; ; )
	{
		op=getkey();
		
		if(op=='+')		accum+=parseval(msflag);
		else if(op=='-')	accum-=parseval(msflag);
		else if(op=='*')	accum*=parseval(msflag);
		else if(op=='/')	accum/=parseval(msflag);
		else if(op=='%')	accum%=parseval(msflag);
		else 			break;
	}
	returnkey(op);
	return accum;
}

static u_int
parseval(char msflag)
{
	KCHAR c, c1;
	u_int temp=0;
	char margoe,  margtype;
	
	c=getkey();
	
	
	/* decimal value */
	
	if(f_isdigit(c))
	{
		do
		{
			temp=mult10(temp)+(c-'0');
		}
		while(f_isdigit(c=getkey()));
		returnkey(c);
		goto to_ms_exit;
	}
	
	/* hex value, ascii contant, tabstop, string length */
	
	switch(f_tolower(c))
	{
		case '$':
			while(f_isxdigit(c=getkey()))
			{
				temp=(temp<<4)+((c>'9') ? f_toupper(c)-'A'+10 : c-'0');
			}
			returnkey(c);
			goto to_ms_exit;
		
		case '#':
			temp=getnumval(getkey());
			goto to_ms_exit;
		
		case '\'':
			temp=getkey();
			goto to_ms_exit;
		
		case '\"':
			temp=strsize();
			goto from_ms_exit;
		
		case 't':
			c1=getkey();
			if(c1>='0' && c1<='9')
			{
				temp=gettabvalue(c1-'0');
				goto from_ms_exit;
			}
			else returnkey(c1);
			break;
			
		default:
			break;
	}
	
	c1=getkey();
	
	dswitch(c, c1)
	{
		dcase('l', 'c'):		/* current line */
			temp=getlnum();
			goto to_ms_exit;
		
		dcase('l', 'p'):		/* lines per page */
			temp=line_page;
			goto to_ms_exit;
		
		dcase('m', 't'):		/* top marg */
			temp=gettopmarg();
			goto to_ms_exit;
		
		dcase('m', 'b'):		/* bot marg */
			temp=getbotmarg();
			goto to_ms_exit;
		
		dcase('o', 'e'):		/* system error value */
			temp=systemerr;
			goto to_ms_exit;
		
		dcase('c', 'o'):		/* number of columns */
			temp=numcols+1;
			goto to_ms_exit;
		
		dcase('i', 'l'):		/* left id value */
			temp=getleftid();
			goto from_ms_exit;
		
		dcase('i', 'r'):		/* right id value */
			temp=getrightid();
			goto from_ms_exit;
		
		dcase('i', 'b'):		 /* both id value */
			temp=getbothid();
			goto from_ms_exit;
		
		dcase('i', 't'):		/* temp id value */
			c=getkey();
			if(f_tolower(c)=='r') temp=gettemprid();
			else
			{
				if(f_tolower(c)!='l') returnkey(c);
				temp=gettempid();
			}
			goto from_ms_exit;
		
		dcase('i', 'd'):		/* delayed id value */
			temp=getdelayid();
			goto from_ms_exit;
		
		dcase('i', 'c'):		/* centered id value */
			temp=getcenterid();
			goto from_ms_exit;
		
		dcase('m', 'l'):		/* left margin */
			whichmarg(&margtype, &margoe);
			temp=getleftmg(margtype, ((f_tolower(margoe)=='o') ? 1 : 0));
			goto from_ms_exit;
		
		dcase('m', 'r'):		/* right margin */
			whichmarg(&margtype, &margoe);
			temp=getrightmg(margtype, ((f_tolower(margoe)=='o') ? 1 : 0));
			goto from_ms_exit;
		
		dcase('m', 'n'):		/* marginal note status */
			temp=mnnotestatus();
			goto to_ms_exit;
		
		dcase('r', 'p'):		/* repeat count */
			temp=getrepeatcount();
			goto to_ms_exit;
		
		dcase('m', 's'):		/* microspace constant */
			temp=msperin;
			goto exit;
		
		dcase('t', 'f'):		/* tab fill char */
			temp=tabfill;
			goto to_ms_exit;
		
		dcase('c', 'c'):		/* control char value */
			temp=control;
			goto to_ms_exit;
		
		dcase('c', 'l'):		/* control line char */
			temp=dot;
			goto to_ms_exit;
		
		dcase('h', 's'):		/* hardspace char */
			temp=hardspace;
			goto to_ms_exit;
		
		dcase('g', 'h'):		/* ghost hyphen char */
			temp=ghosthy;
			goto to_ms_exit;
		
		dcase('j', 'm'):		/* justify mode */
			temp=get_justmode();
			goto to_ms_exit;
		
		dcase('l', 'f'):		/* line format */
			whichmarg(&margtype, &margoe);
			temp=getlineformat(margtype, (f_tolower(margoe)=='o') ? 1 : 0);
			goto to_ms_exit;
			
		default:
			break;
	}
	error("Unknown value/variable '%c%c'", c, c1);
	
to_ms_exit:
	if(msflag) temp*=msperin;
	goto exit;

from_ms_exit:
	if(!msflag) temp/=msperin;
	goto exit;
	
exit:
	return temp;
}

/*	Calculate the size of string in MS. If a font/attribute is encountered
	a temp attribute table will be updated. This means that the size of a
	string will be accurate even with size changes...
*/

static u_int
strsize(void)
{
	KCHAR c;
	u_int temp=0;
	Att_table atts;
	
	/* copy the body attribute table */
	
	memcpy(&atts, getbodyatts(), sizeof(Att_table));
	
	while((c=getkey())!='\"')
	{
		if(!c || c==CR) error("Trailing quote expected");
		
		else if(is_font(c)) atts.cfont=c-INT_FONT;
		
		else if(is_attroff(c))
		{
			c-=INT_ATTROFF;
			atts.aselect[c]=0;
			if(getasize(c)) atts.sizemod=1;
		}
		
		else if(is_attron(c))
		{
			c-=INT_ATTRON;
			atts.aselect[c]=1;
			if(getasize(c)) atts.sizemod=getasize(c);
		}
		else temp+=getcsize(c, &atts);
	}
	return temp;
}

